import contextlib
from typing import (
    ContextManager,
    Generator,
)

import attrs
import pytest
import yt.wrapper as yt

from tests.utils import DOCKER_IMAGE
from tests.yt_instances import YtInstance
from tractorun.backend.generic import GenericBackend
from tractorun.description import (
    DescriptionManager,
    Link,
)
from tractorun.mesh import Mesh
from tractorun.private.constants import (
    TRACTORUN_DESCRIPTION_MANAGER_NAME,
    USER_DESCRIPTION_MANAGER_NAME,
)
from tractorun.private.yt_cluster import (
    TractorunClusterConfig,
    make_cypress_link,
)
from tractorun.run import run
from tractorun.toolbox import Toolbox


@contextlib.contextmanager
def _dummy_context_manager() -> Generator[None, None, None]:
    yield


@pytest.mark.parametrize(
    "config_exists",
    [False, True],
)
def test_description_empty_config(config_exists: bool, yt_path: str, yt_instance: YtInstance) -> None:
    # checking that the basic logic works without config or with an empty config
    yt_client = yt_instance.get_client()

    def checker(toolbox: Toolbox) -> None:
        pass

    config_path = f"{yt_path}/empty_config"

    if config_exists:
        yt_client.create("document", path=config_path)

    mesh = Mesh(node_count=1, process_per_node=1, gpu_per_process=0)
    ctx_manager = (
        pytest.warns(
            UserWarning,
            match=f"Cluster config {config_path} does not exist. Some functions are not available. Please specify config's path by tractorun params.",
        )
        if not config_exists
        else _dummy_context_manager()
    )
    assert isinstance(ctx_manager, ContextManager)
    with ctx_manager:
        run_info = run(
            checker,
            backend=GenericBackend(),
            yt_path=yt_path,
            mesh=mesh,
            yt_client=yt_client,
            docker_image=DOCKER_IMAGE,
            cluster_config_path=config_path,
        )
    operation = yt.Operation(id=run_info.operation_id, client=yt_client)
    description = operation.get_attributes()["runtime_parameters"]["annotations"]["description"]
    tractorun_description = description[TRACTORUN_DESCRIPTION_MANAGER_NAME]
    assert tractorun_description is not None


def test_set_tractorun_description(
    yt_instance: YtInstance, cluster_config: TractorunClusterConfig, cluster_config_path: str, yt_path: str
) -> None:
    assert cluster_config.cypress_link_template is not None
    yt_client = yt_instance.get_client()

    def checker(toolbox: Toolbox) -> None:
        pass

    mesh = Mesh(node_count=1, process_per_node=1, gpu_per_process=0)
    run_info = run(
        checker,
        backend=GenericBackend(),
        yt_path=yt_path,
        mesh=mesh,
        yt_client=yt_client,
        docker_image=DOCKER_IMAGE,
        cluster_config_path=cluster_config_path,
    )
    operation = yt.Operation(id=run_info.operation_id, client=yt_client)
    description = operation.get_attributes()["runtime_parameters"]["annotations"]["description"]
    tractorun_description = description[TRACTORUN_DESCRIPTION_MANAGER_NAME]
    assert str(tractorun_description["training_dir"]) == make_cypress_link(
        path=yt_path,
        cypress_link_template=cluster_config.cypress_link_template,
    )
    assert str(tractorun_description["logs"]) == make_cypress_link(
        path=f"{yt_path}/logs/0",
        cypress_link_template=cluster_config.cypress_link_template,
    )
    assert "primary" in tractorun_description
    assert "job_stderr" in tractorun_description["primary"]
    assert "address" in tractorun_description["primary"]
    assert "job_stderr" in tractorun_description["primary"]
    assert int(tractorun_description["incarnation"]) == 0
    assert tractorun_description["mesh"] == attrs.asdict(mesh)  # type: ignore


def test_set_user_description(yt_instance: YtInstance, cluster_config_path: str, yt_path: str) -> None:
    yt_client = yt_instance.get_client()

    def checker(toolbox: Toolbox) -> None:
        toolbox.description_manager.set(
            {
                "wandb": Link(value="https://fake.wandb.url/some/page"),
                "complex_structure": [
                    Link(value="https://another.com"),
                    [True, 1, 0.5, b"value"],
                    {"cats": "dogs"},
                ],
                "custom_info": "foo",
                "cypress_link": toolbox.description_manager.make_cypress_link("//some/path 1"),
            }
        )

    mesh = Mesh(node_count=1, process_per_node=1, gpu_per_process=0)
    run_info = run(
        checker,
        backend=GenericBackend(),
        yt_path=yt_path,
        mesh=mesh,
        yt_client=yt_client,
        docker_image=DOCKER_IMAGE,
        cluster_config_path=cluster_config_path,
    )
    operation = yt.Operation(id=run_info.operation_id, client=yt_client)
    description = operation.get_attributes()["runtime_parameters"]["annotations"]["description"]
    user_description = description[USER_DESCRIPTION_MANAGER_NAME]
    assert user_description == {
        "wandb": Link(value="https://fake.wandb.url/some/page").to_yson(),
        "complex_structure": [
            Link(value="https://another.com").to_yson(),
            [True, 1, 0.5, "value"],
            {"cats": "dogs"},
        ],
        "custom_info": "foo",
        "cypress_link": Link(value="https://yt.tracto.ai/yt/navigation?path=//some/path%201").to_yson(),
    }


def test_make_description() -> None:
    description = DescriptionManager._make_description(
        key=["1", "2"],
        description={
            "foo": 123,
            "bar": [1, 2, {"1": True}],
            "far": None,
            "link": Link(value=None),
            "bla": Link(value="//foo"),
            "links": [Link(value="link1"), Link(value="link2")],
        },
    )
    assert description == {
        "1": {
            "2": {
                "foo": 123,
                "bar": [1, 2, {"1": True}],
                "far": None,
                "link": Link(value=None).to_yson(),
                "bla": Link(value="//foo").to_yson(),
                "links": [Link(value="link1").to_yson(), Link(value="link2").to_yson()],
            },
        },
    }


def test_convert_yson() -> None:
    converted = DescriptionManager._convert_yson(
        {
            "foo": 123,
            "bar": [1, 2, {"1": True}],
            "far": None,
            "link": Link(value=None),
            "bla": Link(value="//foo"),
            "links": [Link(value="link1"), Link(value="link2")],
            "types": [True, 1, 0.5, b"value"],
        }
    )
    assert converted == {
        "foo": 123,
        "bar": [1, 2, {"1": True}],
        "far": None,
        "link": Link(value=None).to_yson(),
        "bla": Link(value="//foo").to_yson(),
        "links": [Link(value="link1").to_yson(), Link(value="link2").to_yson()],
        "types": [True, 1, 0.5, b"value"],
    }


def test_make_cypress_link(yt_instance: YtInstance) -> None:
    link = DescriptionManager(
        operation_id="123-123",
        cypress_link_template="https://fake.cluster?path={path}",
        yt_client=yt_instance.get_client(),
    ).make_cypress_link("//tmp/some/path")
    assert link == Link(value="https://fake.cluster?path=//tmp/some/path")



import uuid

import pytest
from yt import wrapper as yt

from tests.utils import (
    DOCKER_IMAGE,
    TractoCli,
    get_data_path,
    make_cli_args,
    make_run_config,
    run_config_file,
)
from tests.yt_instances import YtInstance
from tractorun.backend.generic import GenericBackend
from tractorun.cli.tractorun_runner import make_configuration
from tractorun.docker_auth import DockerAuthSecret
from tractorun.exception import DockerAuthDataError
from tractorun.mesh import Mesh
from tractorun.run import run


TEST_SECRET_USER_PASS = {
    "username": "user1",
    "password": "password1",
}
TEST_SECRET_AUTH = {
    "auth": "aksdl",
}


def test_configuration() -> None:
    _, _, config = make_configuration(
        make_cli_args(
            "--docker-auth-secret.cypress-path",
            "//tmp/some_secret_cli",
        )
    )
    assert config.docker_auth_secret == DockerAuthSecret(cypress_path="//tmp/some_secret_cli")

    run_config = make_run_config(
        {
            "docker_auth_secret": {
                "cypress_path": "//tmp/some_secret_config",
            },
        }
    )
    with run_config_file(run_config) as run_config_path:
        _, _, config = make_configuration(["--run-config-path", run_config_path])
    assert config.docker_auth_secret == DockerAuthSecret(cypress_path="//tmp/some_secret_config")

    with run_config_file(run_config) as run_config_path:
        _, _, config = make_configuration(
            ["--run-config-path", run_config_path, "--docker-auth-secret.cypress-path", "//tmp/some_secret_cli"]
        )
    assert config.docker_auth_secret == DockerAuthSecret(cypress_path="//tmp/some_secret_cli")


def create_yt_secret(yt_client: yt.YtClient, secret: dict, yt_path: str) -> str:
    secret_path = yt_path + "/secret"
    yt.create("document", secret_path, client=yt_client)
    yt_client.set(secret_path, secret)
    return secret_path


@pytest.mark.parametrize(
    "secret",
    [
        {
            "username": "user1",
            "password": "password1",
        },
        {
            "auth": "sasdsa",
        },
    ],
)
def test_spec_pickle(secret: dict, yt_instance: YtInstance, yt_path: str) -> None:
    yt_client = yt_instance.get_client()

    yt_secret_path = create_yt_secret(yt_client, secret, yt_path)

    def checker() -> None:
        pass

    operation_title = f"test operation {uuid.uuid4()}"

    run_info = run(
        checker,
        backend=GenericBackend(),
        yt_path=yt_path,
        mesh=Mesh(node_count=1, process_per_node=1, gpu_per_process=0),
        yt_client=yt_client,
        docker_image=DOCKER_IMAGE,
        title=operation_title,
        local=False,
        dry_run=True,
        docker_auth=DockerAuthSecret(cypress_path=yt_secret_path),
    )
    assert run_info.operation_spec["secure_vault"]["docker_auth"] == secret


def test_run_cli(yt_instance: YtInstance, yt_path: str, mnist_ds_path: str) -> None:
    yt_client = yt_instance.get_client()

    secret = {"auth": "123"}
    yt_secret_path = create_yt_secret(yt_client, secret, yt_path)

    tracto_cli = TractoCli(
        command=["python3", "/tractorun_tests/torch_run_script.py"],
        args=[
            "--mesh.node-count",
            "1",
            "--mesh.process-per-node",
            "1",
            "--mesh.gpu-per-process",
            "0",
            "--yt-path",
            yt_path,
            "--bind-local",
            f"{get_data_path('../data/torch_run_script.py')}:/tractorun_tests/torch_run_script.py",
            "--docker-auth-secret.cypress-path",
            yt_secret_path,
        ],
    )
    run_info = tracto_cli.dry_run()
    assert run_info["run_info"]["operation_spec"]["secure_vault"]["docker_auth"] == secret


@pytest.mark.parametrize(
    "secret",
    [
        {
            "username": "user1",
        },
        {
            "password": "user1",
        },
        {
            "auth_data": "sasdsa",
        },
    ],
)
def test_invalid_format(secret: dict, yt_instance: YtInstance, yt_path: str) -> None:
    yt_client = yt_instance.get_client()

    yt_secret_path = create_yt_secret(yt_client, secret, yt_path)

    def checker() -> None:
        pass

    operation_title = f"test operation {uuid.uuid4()}"

    with pytest.raises(DockerAuthDataError):
        _ = run(
            checker,
            backend=GenericBackend(),
            yt_path=yt_path,
            mesh=Mesh(node_count=1, process_per_node=1, gpu_per_process=0),
            yt_client=yt_client,
            docker_image=DOCKER_IMAGE,
            title=operation_title,
            local=False,
            dry_run=True,
            docker_auth=DockerAuthSecret(cypress_path=yt_secret_path),
        )



from collections import Counter
import json
import os
import shutil
from typing import List
import zipfile

import attrs

from tractorun.bind import BindLocal


@attrs.define(kw_only=True, slots=True, auto_attribs=True)
class PackedBind:
    yt_path: str
    local_path: str


@attrs.define(kw_only=True, slots=True, auto_attribs=True)
class BindsPacker:
    _files: List[BindLocal]
    _dirs: List[BindLocal]

    @classmethod
    def from_binds(cls, binds: list[BindLocal]) -> "BindsPacker":
        files: list[BindLocal] = []
        dirs: list[BindLocal] = []

        for bind in sorted(binds, key=lambda x: x.source):
            if os.path.isdir(bind.source):
                dirs.append(bind)
            else:
                files.append(bind)
        return BindsPacker(files=files, dirs=dirs)

    def _pack_file(self, archive_path: str, archive_name: str, source_path: str, destination_name: str) -> PackedBind:
        with zipfile.ZipFile(archive_path + ".zip", "w") as zipf:
            zipf.write(source_path, arcname=destination_name)
        return PackedBind(
            yt_path=f"{archive_name}.zip",
            local_path=f"{archive_path}.zip",
        )

    def _pack_dir(self, archive_path: str, archive_name: str, source_path: str) -> PackedBind:
        shutil.make_archive(archive_path, "zip", source_path)
        return PackedBind(
            yt_path=f"{archive_name}.zip",
            local_path=f"{archive_path}.zip",
        )

    def pack(self, base_result_path: str) -> list[PackedBind]:
        result: list[PackedBind] = []
        for idx, bind in enumerate(self._dirs):
            archive_name = f"__dir_{idx}"
            path = os.path.join(base_result_path, archive_name)
            packed_bind = self._pack_dir(
                archive_path=path,
                archive_name=archive_name,
                source_path=bind.source,
            )
            result.append(packed_bind)
        for idx, bind in enumerate(self._files):
            archive_name = f"__file_{idx}"
            path = os.path.join(base_result_path, archive_name)
            destination_name = os.path.split(bind.destination)[1]
            packed_bind = self._pack_file(
                archive_path=path,
                archive_name=archive_name,
                source_path=bind.source,
                destination_name=destination_name,
            )
            result.append(packed_bind)
        return result

    def unpack(self) -> None:
        # Runs inside a job
        for idx, bind in enumerate(self._dirs):
            path_w = f"__dir_{idx}.zip"
            with zipfile.ZipFile(path_w, "r") as zipf:
                zipf.extractall(path=bind.destination)
        for idx, bind in enumerate(self._files):
            path_w = f"__file_{idx}.zip"
            destination = os.path.split(bind.destination)[0]
            with zipfile.ZipFile(path_w, "r") as zipf:
                zipf.extractall(path=destination)

    def to_env(self) -> str:
        return json.dumps(
            {
                "files": [attrs.asdict(bind) for bind in self._files],  # type: ignore
                "dirs": [attrs.asdict(bind) for bind in self._dirs],  # type: ignore
            },
        )

    @classmethod
    def from_env(cls, content: str) -> "BindsPacker":
        parsed = json.loads(content)
        files = [BindLocal(source=record["source"], destination=record["destination"]) for record in parsed["files"]]
        dirs = [BindLocal(source=record["source"], destination=record["destination"]) for record in parsed["dirs"]]
        return BindsPacker(files=files, dirs=dirs)


@attrs.define(kw_only=True, slots=True, auto_attribs=True)
class PackedLib:
    archive_name: str
    path: str


@attrs.define(kw_only=True, slots=True, auto_attribs=True)
class BindsLibPacker:
    _paths: list[str]

    def pack(self, base_result_path: str) -> list[PackedLib]:
        paths = [os.path.abspath(path) for path in self._paths]
        lib_to_original_path = {os.path.split(path)[1]: path for path in paths}
        non_uniq_names = [name for name, count in Counter(lib_to_original_path.keys()).items() if count > 1]
        assert len(non_uniq_names) == 0, f"Some libraries have the same names: {non_uniq_names}"

        result: list[PackedLib] = []
        for lib_name, original_path in lib_to_original_path.items():
            path = os.path.join(base_result_path, f"__{lib_name}")
            root_dir, base_dir = os.path.split(os.path.abspath(original_path))
            shutil.make_archive(path, "zip", root_dir, base_dir)
            result.append(
                PackedLib(
                    archive_name=lib_name,
                    path=f"{path}.zip",
                )
            )
        return result



import base64
import io
import json
import os
import pickle
import subprocess
from typing import (
    TYPE_CHECKING,
    NewType,
)

import attrs

from tractorun.mesh import Mesh
from tractorun.private.constants import TRACTO_CONFIG_ENV_VAR
from tractorun.private.helpers import AttrSerializer
from tractorun.private.training_dir import TrainingDir
from tractorun.private.yt_cluster import TractorunClusterConfig


WorkerIndex = NewType("WorkerIndex", int)


@attrs.define(kw_only=True, slots=True, auto_attribs=True)
class WorkerConfig:
    mesh: Mesh
    port: int
    node_index: int
    proc_index: int
    self_index: int
    training_dir: TrainingDir
    yt_client_config: str
    operation_id: str
    job_id: str
    cluster_config: TractorunClusterConfig


@attrs.define(kw_only=True, slots=True, auto_attribs=True)
class WorkerRun:
    worker_config: WorkerConfig
    _process: subprocess.Popen

    @classmethod
    def run(
        cls,
        command: list[str],
        mesh: Mesh,
        self_index: int,
        node_index: int,
        proc_index: int,
        port: int,
        training_dir: TrainingDir,
        cluster_config: TractorunClusterConfig,
        yt_config: dict,
        env: dict,
    ) -> "WorkerRun":
        worker_config = WorkerConfig(
            mesh=mesh,
            node_index=node_index,
            proc_index=proc_index,
            self_index=self_index,
            port=port,
            training_dir=training_dir,
            yt_client_config=base64.b64encode(pickle.dumps(yt_config)).decode("utf-8"),
            cluster_config=cluster_config,
            operation_id=env["YT_OPERATION_ID"],
            job_id=env["YT_JOB_ID"],
        )
        config_name = f"config_{proc_index}.json"
        with open(config_name, "w") as f:
            serializer = AttrSerializer(WorkerConfig)
            json.dump(serializer.serialize(worker_config), f)

        # TODO: torch multiprocessing is better (or another backend-specific tool),
        # but pickling does not work in this case.
        # torch.multiprocessing.spawn(wrapped_run_mp, nprocs=mesh.process_per_node, args=(f, c, path,), join=True)
        process = subprocess.Popen(
            command,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            bufsize=1,
            universal_newlines=True,
            env={
                TRACTO_CONFIG_ENV_VAR: config_name,
                "YT_PROXY": yt_config["proxy"]["url"],
                "YT_TOKEN": yt_config["token"],
                **env,
            },
        )
        if TYPE_CHECKING:
            assert isinstance(process.stdout, io.TextIOWrapper)
            assert isinstance(process.stderr, io.TextIOWrapper)
        os.set_blocking(process.stdout.fileno(), False)
        os.set_blocking(process.stderr.fileno(), False)
        return WorkerRun(
            worker_config=worker_config,
            process=process,
        )

    def poll(self) -> int | None:
        return self._process.poll()

    def wait(self) -> None:
        self._process.wait()

    def terminate(self) -> None:
        self._process.terminate()

    def stdout(self) -> io.TextIOWrapper:
        if TYPE_CHECKING:
            assert isinstance(self._process.stdout, io.TextIOWrapper)
        return self._process.stdout

    def stderr(self) -> io.TextIOWrapper:
        if TYPE_CHECKING:
            assert isinstance(self._process.stderr, io.TextIOWrapper)
        return self._process.stderr



import attrs

from tractorun.private.helpers import AttrSerializer as _AttrSerializer


__all__ = ["EnvVariable"]


@attrs.define(kw_only=True, slots=True, auto_attribs=True)
class EnvVariable:
    name: str
    value: str | None = None
    cypress_path: str | None = None

    @staticmethod
    def from_args(value: str) -> "EnvVariable":
        return _AttrSerializer(EnvVariable).deserialize(value)



import os

import attrs


__all__ = ["BindLocal", "BindCypress"]


def _to_abs_path(path: str) -> str:
    # mypy workaround
    return os.path.abspath(path)


@attrs.define(kw_only=True, slots=True)
class BindLocal:
    # TODO: just use pathlib
    source: str = attrs.field(converter=_to_abs_path)
    destination: str = attrs.field(converter=_to_abs_path)


@attrs.define(kw_only=True, slots=True)
class BindCypress:
    source: str = attrs.field()
    destination: str = attrs.field()



import os

from tractorun.description import Link
from tractorun.private.closet import Closet
from tractorun.private.constants import TRACTORUN_DESCRIPTION_MANAGER_NAME
from tractorun.private.description import TractorunDescription
from tractorun.private.yt_cluster import (
    make_cypress_link,
    make_job_stderr_link,
)
from tractorun.toolbox import Toolbox


def get_toolbox(closet: Closet) -> Toolbox:
    user_description_manager = closet.description_manager.get_child("extra")
    toolbox = Toolbox(
        coordinator=closet.coordinator,
        checkpoint_manager=closet.checkpoint_manager,
        yt_client=closet.yt_client,
        mesh=closet.mesh,
        training_dir=closet.training_dir,
        training_metadata=closet.training_metadata,
        description_manager=user_description_manager,
    )

    return toolbox


def prepare_log_dir(closet: Closet) -> None:
    logs_dir = f"{closet.training_dir.logs_path}/{closet.coordinator.get_incarnation_id()}"
    worker_logs_dir = f"{logs_dir}/workers"
    sidecar_logs_dir = f"{logs_dir}/sidecars"
    closet.yt_client.link(
        closet.training_dir.worker_logs_path,
        worker_logs_dir,
        recursive=True,
    )
    closet.yt_client.link(
        closet.training_dir.sidecar_logs_path,
        sidecar_logs_dir,
        recursive=True,
    )


def make_description(closet: Closet) -> TractorunDescription:
    logs_dir = f"{closet.training_dir.logs_path}/{closet.coordinator.get_incarnation_id()}"
    training_dir = make_cypress_link(
        path=closet.training_dir.base_path,
        cypress_link_template=closet.cluster_config.cypress_link_template,
    )
    job_stderr_link = make_job_stderr_link(
        operation_id=closet.training_metadata.operation_id,
        job_id=closet.training_metadata.job_id,
        job_stderr_link_template=closet.cluster_config.job_stderr_link_template,
    )
    logs_path = make_cypress_link(
        path=f"{logs_dir}",
        cypress_link_template=closet.cluster_config.cypress_link_template,
    )
    return TractorunDescription(
        training_dir=Link(value=training_dir),
        primary_address=closet.coordinator.get_primary_endpoint(),
        incarnation=closet.coordinator.get_incarnation_id(),
        mesh=closet.mesh,
        primary_stderr=Link(value=job_stderr_link),
        logs=Link(value=logs_path),
    )


def prepare_environment(closet: Closet) -> None:
    # Runs in a job
    ep = closet.coordinator.get_primary_endpoint()
    os.environ["MASTER_ADDR"] = ep.split(":")[0]
    os.environ["MASTER_PORT"] = ep.split(":")[1]
    os.environ["WORLD_SIZE"] = str(closet.coordinator.get_total_peer_count())
    os.environ["NODE_RANK"] = str(closet.coordinator.get_self_index() // closet.mesh.process_per_node)
    os.environ["LOCAL_RANK"] = str(closet.coordinator.get_self_index() % closet.mesh.process_per_node)

    if closet.coordinator.is_primary():
        prepare_log_dir(closet=closet)
        description_manager = closet.description_manager.get_child(TRACTORUN_DESCRIPTION_MANAGER_NAME)
        description = make_description(closet=closet)
        description_manager.set(description.to_dict())



import os
import shutil
from typing import Optional

import attrs
from yt import yson

from tractorun.sidecar import (
    RestartPolicy,
    Sidecar,
)
from tractorun.tensorproxy import TensorproxySidecar


@attrs.define(kw_only=True, slots=True, auto_attribs=True)
class TensorproxyBootstrap:
    restart_policy: RestartPolicy
    ports_count: int

    def prepare_and_get_sidecars(self, yt_proxy: str, grpc_port: int, monitoring_port: int) -> list[Sidecar]:
        path = "__tensorproxy_config.yson"
        with open(path, "wb") as f:
            config_content = get_config(proxy=yt_proxy, grpc_port=grpc_port, monitoring_port=monitoring_port)
            f.write(yson.dumps(config_content))
        shutil.copy("./tensorproxy", "./tensorproxy_fix_perm")
        os.chmod("./tensorproxy_fix_perm", 0o755)
        # TODO: customize binary's name
        return [
            Sidecar(command=["./tensorproxy_fix_perm", "--config", path], restart_policy=self.restart_policy),
        ]

    def get_environment(self, grpc_port: int) -> dict:
        return {
            "TS_GRPC_ADDRESS": f"localhost:{grpc_port}",
        }


@attrs.define(kw_only=True, slots=True, auto_attribs=True)
class TensorproxyConfigurator:
    _tensorproxy: Optional[TensorproxySidecar]

    def generate_configuration(self) -> tuple[Optional[TensorproxyBootstrap], list[str], int]:
        if self._tensorproxy is None or self._tensorproxy.enabled is False:
            return None, [], 0
        ports_count = 2  # 1 for monitoring and 1 for grpc handler
        return (
            TensorproxyBootstrap(
                restart_policy=self._tensorproxy.restart_policy,
                ports_count=ports_count,
            ),
            [self._tensorproxy.yt_path],
            ports_count,
        )


def get_config(grpc_port: int, monitoring_port: int, proxy: str) -> dict:
    return {
        "monitoring_port": monitoring_port,
        "grpc_port": grpc_port,
        "store_kind": "cypress",
        "cypress_store": {"cluster_url": proxy, "file_writer": {"block_size": 262144}},
        "logging": {
            "flush_period": 100,
            "rules": [
                {"exclude_categories": ["Bus"], "family": "plain_text", "min_level": "debug", "writers": ["debug"]},
                {"family": "plain_text", "min_level": "info", "writers": ["info"]},
                {"family": "plain_text", "min_level": "error", "writers": ["error"]},
            ],
            "writers": {
                "debug": {
                    "enable_system_messages": True,
                    "file_name": "./tensorproxy.debug.log",
                    "format": "plain_text",
                    "rotation_policy": {
                        "max_segment_count_to_keep": 1000,
                        "max_total_size_to_keep": 130000000000,
                        "rotation_period": 900000,
                    },
                    "type": "file",
                },
                "error": {"enable_system_messages": True, "format": "plain_text", "type": "stderr"},
                "info": {
                    "enable_system_messages": True,
                    "file_name": "./tensorproxy.info.log",
                    "format": "plain_text",
                    "rotation_policy": {
                        "max_segment_count_to_keep": 1000,
                        "max_total_size_to_keep": 130000000000,
                        "rotation_period": 900000,
                    },
                    "type": "file",
                },
            },
        },
    }



import pytest

from tests.utils import run_config_file
from tractorun.cli.tractorun_runner import make_configuration
from tractorun.exception import TractorunConfigError


def test_fail_unknown_key() -> None:
    with run_config_file({"__SOME_UNKNOWN_KEY__": []}) as run_config_path:
        with pytest.raises(TractorunConfigError):
            _, _, config = make_configuration(["--run-config-path", run_config_path])



import abc
import base64
import copy
import json
import os
import pickle
import random
import shlex
import sys
import tempfile
from typing import (
    Any,
    Callable,
    Optional,
)

import attrs
import cattrs.errors
from yt import wrapper as yt
from yt.wrapper import TaskSpecBuilder
import yt.yson as yson

from tractorun import __version__
from tractorun.base_backend import BackendBase
from tractorun.bind import (
    BindCypress,
    BindLocal,
)
from tractorun.docker_auth import DockerAuthData
from tractorun.env import EnvVariable
from tractorun.exception import (
    TractorunConfigurationError,
    TractorunVersionMismatchError,
)
from tractorun.mesh import Mesh
from tractorun.operation_log import OperationLogMode
from tractorun.private import constants as const
from tractorun.private.bind import (
    BindsLibPacker,
    BindsPacker,
)
from tractorun.private.bootstrapper import (
    BootstrapConfig,
    LibVersions,
    bootstrap,
)
from tractorun.private.closet import get_closet
from tractorun.private.constants import (
    BIND_PATHS_ENV_VAR,
    BOOTSTRAP_CONFIG_FILENAME_ENV_VAR,
    BOOTSTRAP_CONFIG_NAME,
)
from tractorun.private.coordinator import get_incarnation_id
from tractorun.private.docker_auth import DockerAuthDataExtractor
from tractorun.private.environment import get_toolbox
from tractorun.private.helpers import (
    AttrSerializer,
    create_attrs_converter,
)
from tractorun.private.stderr_reader import StderrReaderWorker
from tractorun.private.tensorproxy import (
    TensorproxyBootstrap,
    TensorproxyConfigurator,
)
from tractorun.private.training_dir import (
    TrainingDir,
    prepare_training_dir,
)
from tractorun.private.yt_cluster import TractorunClusterConfig
from tractorun.resources import Resources
from tractorun.run_info import RunInfo
from tractorun.sidecar import Sidecar
from tractorun.stderr_reader import StderrMode
from tractorun.tensorproxy import TensorproxySidecar
from tractorun.toolbox import Toolbox


class Runnable(abc.ABC):
    @abc.abstractmethod
    def modify_task(self, task: yt.TaskSpecBuilder) -> yt.TaskSpecBuilder:
        pass

    @abc.abstractmethod
    def modify_operation(self, operation: yt.VanillaSpecBuilder) -> yt.VanillaSpecBuilder:
        pass

    @abc.abstractmethod
    def get_bootstrap_command(self) -> list[str]:
        pass

    @abc.abstractmethod
    def make_yt_command(self) -> Callable | bytes:
        pass

    @abc.abstractmethod
    def make_local_command(
        self,
        mesh: Mesh,
        sidecars: list[Sidecar],
        env: list[EnvVariable],
        training_dir: TrainingDir,
        yt_client_config: str,
        tensorproxy: Optional[TensorproxyBootstrap],
        lib_versions: LibVersions,
        cluster_config: TractorunClusterConfig,
        operation_log_mode: OperationLogMode,
    ) -> Callable:
        pass


@attrs.define
class CliCommand(Runnable):
    command: list[str]

    def modify_task(self, task: yt.TaskSpecBuilder) -> yt.TaskSpecBuilder:
        return task

    def modify_operation(self, operation: yt.VanillaSpecBuilder) -> yt.VanillaSpecBuilder:
        return operation

    def get_bootstrap_command(self) -> list[str]:
        return self.command

    def make_yt_command(self) -> bytes:
        escaped_command = " ".join([shlex.quote(arg) for arg in self.command])
        return f"python3 -m tractorun.cli.tractorun_bootstrap {escaped_command}".encode("utf-8")

    def make_local_command(
        self,
        mesh: Mesh,
        sidecars: list[Sidecar],
        env: list[EnvVariable],
        training_dir: TrainingDir,
        yt_client_config: str,
        tensorproxy: Optional[TensorproxyBootstrap],
        lib_versions: LibVersions,
        cluster_config: TractorunClusterConfig,
        operation_log_mode: OperationLogMode,
    ) -> Callable:
        def wrapped() -> None:
            bootstrap(
                mesh=mesh,
                training_dir=training_dir,
                yt_client_config=yt_client_config,
                command=self.get_bootstrap_command(),
                sidecars=sidecars,
                env=env,
                tensorproxy=tensorproxy,
                lib_versions=lib_versions,
                cluster_config=cluster_config,
                operation_log_mode=operation_log_mode,
            )

        return wrapped


@attrs.define
class UserFunction(Runnable):
    function: Callable
    _backend: BackendBase

    def modify_task(self, task: yt.TaskSpecBuilder) -> yt.TaskSpecBuilder:
        return task

    def modify_operation(self, operation: yt.VanillaSpecBuilder) -> yt.VanillaSpecBuilder:
        return operation

    def get_bootstrap_command(self) -> list[str]:
        # on YT it should be native YT-wrapper command like
        # python3 _py_runner.py wrapped.pickle config_dump _modules_info _main_module.py _main_module PY_SOURCE
        return ["python3"] + sys.argv

    def make_yt_command(self) -> Callable:
        def wrapped() -> None:
            # run on YT
            if "TRACTO_CONFIG" in os.environ:
                toolbox = prepare_and_get_toolbox(backend=self._backend)
                self.function(toolbox)
            else:
                binds_packer = BindsPacker.from_env(os.environ[BIND_PATHS_ENV_VAR])
                binds_packer.unpack()
                bootstrap_config_path = os.environ[BOOTSTRAP_CONFIG_FILENAME_ENV_VAR]
                with open(bootstrap_config_path, "r") as f:
                    content = f.read()
                    deserializer = AttrSerializer(
                        BootstrapConfig,
                        # forward compatibility
                        converter=create_attrs_converter(forbid_extra_keys=False),
                    )
                    try:
                        config: BootstrapConfig = deserializer.deserialize(data=content)
                    except cattrs.errors.BaseValidationError as e:
                        raise TractorunVersionMismatchError(
                            "Please check that the tractorun version locally and on YT are the same",
                        ) from e
                bootstrap(
                    mesh=config.mesh,
                    training_dir=config.training_dir,
                    yt_client_config=config.yt_client_config,
                    command=self.get_bootstrap_command(),
                    sidecars=config.sidecars,
                    env=config.env,
                    tensorproxy=config.tensorproxy,
                    lib_versions=config.lib_versions,
                    cluster_config=config.cluster_config,
                    operation_log_mode=config.operation_log_mode,
                )

        return wrapped

    def make_local_command(
        self,
        mesh: Mesh,
        sidecars: list[Sidecar],
        env: list[EnvVariable],
        training_dir: TrainingDir,
        yt_client_config: str,
        tensorproxy: Optional[TensorproxyBootstrap],
        lib_versions: LibVersions,
        cluster_config: TractorunClusterConfig,
        operation_log_mode: OperationLogMode,
    ) -> Callable:
        def wrapped() -> None:
            # run on YT
            if "TRACTO_CONFIG" in os.environ:
                toolbox = prepare_and_get_toolbox(backend=self._backend)
                self.function(toolbox)
            else:
                bootstrap(
                    mesh=mesh,
                    training_dir=training_dir,
                    yt_client_config=yt_client_config,
                    command=self.get_bootstrap_command(),
                    sidecars=sidecars,
                    env=env,
                    tensorproxy=tensorproxy,
                    lib_versions=lib_versions,
                    cluster_config=cluster_config,
                    operation_log_mode=operation_log_mode,
                )

        return wrapped


@attrs.define(kw_only=True, slots=True, auto_attribs=True)
class TractorunParams:
    runnable: Runnable
    docker_image: str
    yt_path: str
    mesh: Mesh
    proxy_stderr_mode: StderrMode
    operation_log_mode: OperationLogMode
    cluster_config_path: str
    title: str | None = None
    user_config: dict[Any, Any]
    binds_local: list[BindLocal]
    binds_local_lib: list[str]
    binds_cypress: list[BindCypress]
    tensorproxy: TensorproxySidecar | None
    no_wait: bool
    sidecars: list[Sidecar]
    env: list[EnvVariable]
    resources: Resources
    yt_client: yt.YtClient | None
    yt_operation_spec: dict[Any, Any]
    yt_task_spec: dict[Any, Any]
    docker_auth: DockerAuthData | None
    attach_external_libs: bool
    dry_run: bool


def run_tracto(params: TractorunParams) -> RunInfo:
    # if mesh.node_count > 1 and mesh.gpu_per_process * mesh.process_per_node not in (0, 8):
    #     raise exc.TractorunInvalidConfiguration("gpu per node can only be 0 or 8")

    yt_client = params.yt_client or yt.YtClient(config=yt.default_config.get_config_from_env())
    yt_client.config["pickling"]["ignore_system_modules"] = False if params.attach_external_libs else True

    # we store it explicitly since locally it could have been read from ~/.yt/token
    yt_client.config["token"] = yt.http_helpers.get_token(client=yt_client) or ""

    yt_client_config = yt.config.get_config(yt_client)
    yt_client_config_pickled = base64.b64encode(pickle.dumps(yt_client_config)).decode("utf-8")

    yt_client_config_for_job: dict = copy.deepcopy(yt_client_config)

    # for tests only
    yt_config_for_job_patch_yson_string = os.environ.get("TRACTORUN_YT_CONFIG_FOR_JOB_PATCH")
    if yt_config_for_job_patch_yson_string:
        patch = yt.yson.loads(yt_config_for_job_patch_yson_string.encode())
        yt.common.update_inplace(yt_client_config_for_job, patch)

    yt_client_config_for_job_pickled = base64.b64encode(pickle.dumps(yt_client_config_for_job)).decode("utf-8")

    # detached mode only applies to the local client
    yt_client.config["detached"] = True if params.no_wait else False

    tmp_dir = tempfile.TemporaryDirectory()
    training_dir = TrainingDir.create(params.yt_path)

    tp_bootstrap, tp_yt_files, tp_ports = TensorproxyConfigurator(
        tensorproxy=params.tensorproxy
    ).generate_configuration()

    cluster_config = TractorunClusterConfig.load_from_yt(yt_client=yt_client, path=params.cluster_config_path)

    bootstrap_config = BootstrapConfig(
        mesh=params.mesh,
        sidecars=params.sidecars,
        env=params.env,
        training_dir=training_dir,
        yt_client_config=yt_client_config_for_job_pickled,
        tensorproxy=tp_bootstrap,
        lib_versions=LibVersions.create(),
        cluster_config=cluster_config,
        operation_log_mode=params.operation_log_mode,
    )

    bootstrap_config_path = os.path.join(tmp_dir.name, BOOTSTRAP_CONFIG_NAME)
    with open(bootstrap_config_path, "w") as f:
        f.write(AttrSerializer(BootstrapConfig).serialize(bootstrap_config))

    binds_packer = BindsPacker.from_binds(
        binds=params.binds_local,
    )
    packed_binds = binds_packer.pack(tmp_dir.name)
    bind_libs_packer = BindsLibPacker(
        paths=params.binds_local_lib,
    )
    packed_libs = bind_libs_packer.pack(tmp_dir.name)
    new_pythonpath = ":".join(["./" + packed_lib.archive_name for packed_lib in packed_libs])

    yt_file_bindings = []
    yt_file_bindings.extend(
        [yt.LocalFile(packed_bind.local_path, packed_bind.yt_path) for packed_bind in packed_binds],
    )

    yt_file_bindings.extend(
        [yson.to_yson_type(cb.source, attributes={"file_name": cb.destination}) for cb in params.binds_cypress]
    )

    yt_file_bindings.extend([yt.LocalFile(packed_lib.path, packed_lib.archive_name) for packed_lib in packed_libs])
    yt_file_bindings.append(
        yt.LocalFile(bootstrap_config_path, BOOTSTRAP_CONFIG_NAME),
    )

    yt_command = params.runnable.make_yt_command()

    # prepare task spec
    task_spec: TaskSpecBuilder = yt.VanillaSpecBuilder().begin_task("task")

    task_spec = task_spec.file_paths(yt_file_bindings + tp_yt_files)

    task_spec = params.runnable.modify_task(
        task_spec.command(yt_command)
        .job_count(params.mesh.node_count)
        .gpu_limit(params.mesh.gpu_per_process * params.mesh.process_per_node)
        .port_count(params.mesh.process_per_node + tp_ports)
        .cpu_limit(params.resources.cpu_limit)
        .memory_limit(params.resources.memory_limit)
        .docker_image(params.docker_image)
        .spec(params.yt_task_spec)
        .environment(
            {
                "YT_ALLOW_HTTP_REQUESTS_TO_YT_FROM_JOB": "1",
                const.YT_USER_CONFIG_ENV_VAR: json.dumps(params.user_config),
                # Sometimes we can't read compiled bytecode in forks on yt.
                "PYTHONDONTWRITEBYTECODE": "1",
                BIND_PATHS_ENV_VAR: binds_packer.to_env(),
                "PYTHONPATH": f"$PYTHONPATH:{new_pythonpath}" if new_pythonpath else "$PYTHONPATH",
                BOOTSTRAP_CONFIG_FILENAME_ENV_VAR: BOOTSTRAP_CONFIG_NAME,
            },
        )
    )

    # prepare operation spec
    operation_spec = task_spec.end_task()
    additional_operation_spec = copy.deepcopy(params.yt_operation_spec)

    operation_spec = operation_spec.title(params.title)

    if params.mesh.pool_trees is not None:
        operation_spec = operation_spec.pool_trees(params.mesh.pool_trees)
    if params.mesh.pool is not None:
        operation_spec = operation_spec.pool(params.mesh.pool)

    secure_vault: dict[str, Any] = {}
    if params.docker_auth:
        secure_vault["docker_auth"] = DockerAuthDataExtractor(yt_client=yt_client).extract(params.docker_auth).to_spec()
        operation_spec.secure_vault(secure_vault)

    # "fail_on_job_restart" is useful for gang operations,
    # so let's turn it on unless disabled explicitly.
    if "fail_on_job_restart" not in additional_operation_spec:
        additional_operation_spec["fail_on_job_restart"] = True

    if "is_gang" not in additional_operation_spec:
        additional_operation_spec["is_gang"] = True

    additional_operation_spec["annotations"] = additional_operation_spec.get("annotations", {})
    additional_operation_spec["annotations"]["is_tractorun"] = True

    # save job stderr for 150 jobs
    # we can't make it bigger because 150 is a scheduler's limit
    operation_spec.max_stderr_count(150)

    operation_spec = operation_spec.spec(additional_operation_spec)
    operation_spec = params.runnable.modify_operation(operation_spec)

    prev_incarnation_id = get_incarnation_id(yt_client, training_dir)

    operation_id = None
    is_sync = not params.no_wait
    if not params.dry_run:
        prepare_training_dir(yt_client=yt_client, training_dir=training_dir)
        with StderrReaderWorker(
            prev_incarnation_id=prev_incarnation_id,
            training_dir=training_dir,
            yt_client_config_pickled=yt_client_config_pickled,
            mode=params.proxy_stderr_mode,
            mesh=params.mesh,
        ):
            operation = yt_client.run_operation(operation_spec, sync=is_sync)
            assert isinstance(operation, yt.Operation)
            operation_id = operation.id

    run_info = RunInfo(
        operation_spec=operation_spec.build(client=yt_client),
        operation_id=operation_id,
    )

    tmp_dir.cleanup()

    return run_info


def run_local(
    params: TractorunParams,
) -> RunInfo:
    if params.mesh.node_count != 1:
        raise TractorunConfigurationError("local mode only supports 1 node")

    yt_client = params.yt_client or yt.YtClient(config=yt.default_config.get_config_from_env())

    # TODO: respawn in docker
    # Fake YT job environment.
    os.environ["YT_OPERATION_ID"] = "1-2-3-4"
    os.environ["YT_JOB_ID"] = "a-b-c-d"
    os.environ["YT_JOB_COOKIE"] = "0"
    os.environ["YT_ALLOW_HTTP_REQUESTS_TO_YT_FROM_JOB"] = "1"

    cluster_config = TractorunClusterConfig.load_from_yt(yt_client=yt_client, path=params.cluster_config_path)

    # TODO: look for free ports
    start_port = random.randint(10000, 20000)
    for i in range(params.mesh.process_per_node):
        os.environ[f"YT_PORT_{i}"] = str(start_port + i)

    training_dir = TrainingDir.create(params.yt_path)
    tp_bootstrap, _, _ = TensorproxyConfigurator(tensorproxy=params.tensorproxy).generate_configuration()

    wrapped = params.runnable.make_local_command(
        mesh=params.mesh,
        sidecars=params.sidecars,
        env=params.env or [],
        training_dir=training_dir,
        yt_client_config=base64.b64encode(pickle.dumps(yt_client.config)).decode("utf-8"),
        tensorproxy=tp_bootstrap,
        lib_versions=LibVersions(
            tractorun=__version__,
            ytsaurus_client=yt.__version__,
        ),
        cluster_config=cluster_config,
        operation_log_mode=params.operation_log_mode,
    )
    if not params.dry_run:
        prepare_training_dir(training_dir, yt_client)
        wrapped()
    return RunInfo(operation_spec={}, operation_id=None)


def prepare_and_get_toolbox(backend: BackendBase) -> Toolbox:
    # Runs in a job
    closet = get_closet()
    backend.environment.prepare(closet)
    return get_toolbox(closet)



import torch
import torch.distributed

from tractorun.base_backend import EnvironmentBase
from tractorun.private.closet import Closet as _Closet
from tractorun.private.environment import prepare_environment as _common_prepare_environment


__all__ = ["Environment"]


class Environment(EnvironmentBase):
    def prepare(self, closet: _Closet) -> None:
        _common_prepare_environment(closet)

        if closet.mesh.gpu_per_process > 0:
            assert torch.cuda.is_available()

            if closet.mesh.gpu_per_process > 1:
                raise RuntimeError("not supported")

            device_index = closet.coordinator.get_process_index()
            assert device_index < torch.cuda.device_count()
            torch.cuda.set_device(closet.coordinator.get_process_index())

        backend = "gloo" if closet.mesh.gpu_per_process == 0 else "nccl"
        torch.distributed.init_process_group(
            backend=backend,
            init_method="tcp://" + closet.coordinator.get_primary_endpoint(),
            rank=closet.coordinator.get_self_index(),
            world_size=closet.coordinator.get_total_peer_count(),
        )



import base64
import contextlib
import datetime
import enum
import pickle
import selectors
import sys
from typing import (
    TYPE_CHECKING,
    Generator,
    Mapping,
    Optional,
)

import attrs
from yt.common import update_inplace

from tractorun.mesh import Mesh
from tractorun.private.operation_log import (
    LogHandler,
    LogHandlerFactory,
    LogRecord,
)
from tractorun.private.sidecar import (
    RestartVerdict,
    SidecarIndex,
    SidecarRun,
)
from tractorun.private.training_dir import TrainingDir
from tractorun.private.worker import (
    WorkerIndex,
    WorkerRun,
)
from tractorun.private.yt_cluster import TractorunClusterConfig
from tractorun.sidecar import Sidecar


class OutputType(str, enum.Enum):
    stdout: str = "stdout"
    stderr: str = "stderr"


class ProcessManagerPollStatus(enum.IntEnum):
    success: int = 0
    fail: int = 1
    running: int = -1


@attrs.define(kw_only=True, slots=True, auto_attribs=True)
class SelectorMeta:
    log_handlers: list[LogHandler]
    output_type: OutputType


@attrs.define(kw_only=True, slots=True, auto_attribs=True)
class SidecarRunMeta:
    sidecar_run: SidecarRun
    log_handlers: list[LogHandler]


@attrs.define(kw_only=True, slots=True, auto_attribs=True)
class ProcessManager:
    _sidecar_runs: dict[SidecarIndex, SidecarRunMeta]
    _worker_runs: dict[WorkerIndex, WorkerRun]
    _io_selector: selectors.DefaultSelector
    _log_handlers: list[LogHandler]

    @classmethod
    @contextlib.contextmanager
    def start(
        cls,
        command: list[str],
        sidecars: list[Sidecar],
        training_dir: TrainingDir,
        yt_client_config: str,
        cluster_config: TractorunClusterConfig,
        mesh: Mesh,
        node_index: int,
        os_environ: Mapping,
        tp_env: dict,
        spec_env: dict,
        log_handler_factories: list[LogHandlerFactory],
    ) -> Generator["ProcessManager", None, None]:
        pm = ProcessManager._start(
            command=command,
            sidecars=sidecars,
            training_dir=training_dir,
            yt_client_config=yt_client_config,
            cluster_config=cluster_config,
            mesh=mesh,
            node_index=node_index,
            os_environ=os_environ,
            tp_env=tp_env,
            spec_env=spec_env,
            log_handler_factories=log_handler_factories,
        )
        yield pm
        pm._stop()

    @classmethod
    def _start(
        cls,
        command: list[str],
        sidecars: list[Sidecar],
        training_dir: TrainingDir,
        yt_client_config: str,
        cluster_config: TractorunClusterConfig,
        mesh: Mesh,
        node_index: int,
        os_environ: Mapping,
        tp_env: dict,
        spec_env: dict,
        log_handler_factories: list[LogHandlerFactory],
    ) -> "ProcessManager":
        worker_runs: dict[WorkerIndex, WorkerRun] = {}
        sidecar_runs: dict[SidecarIndex, SidecarRun] = {}
        selector = selectors.DefaultSelector()

        yt_config = pickle.loads(base64.b64decode(yt_client_config))
        update_inplace(
            yt_config,
            {
                "pickling": {
                    "module_filter": None,
                },
            },
        )

        for proc_index in range(mesh.process_per_node):
            port = int(os_environ[f"YT_PORT_{proc_index}"])
            self_index = node_index * mesh.process_per_node + proc_index
            worker_run = WorkerRun.run(
                command=command,
                mesh=mesh,
                self_index=self_index,
                node_index=node_index,
                proc_index=proc_index,
                port=port,
                training_dir=training_dir,
                cluster_config=cluster_config,
                yt_config=yt_config,
                env={
                    **os_environ,
                    **tp_env,
                    **spec_env,
                },
            )
            worker_index = WorkerIndex(self_index)
            worker_runs[worker_index] = worker_run

        for local_index, sidecar in enumerate(sidecars):
            sidecar_index = SidecarIndex(node_index * len(sidecars) + local_index)
            sidecar_run = SidecarRun.run(
                sidecar=sidecar,
                env={
                    **os_environ,
                    "YT_PROXY": yt_config["proxy"]["url"],
                    "YT_TOKEN": yt_config["token"],
                    **spec_env,  # TODO(gritukan): Make separate env for sidecars
                },
            )
            sidecar_runs[sidecar_index] = sidecar_run

        log_handlers = []

        for worker_index, worker_run in worker_runs.items():
            worker_log_handlers = []
            for factory in log_handler_factories:
                handler = factory.create_for_worker(worker_index, worker_run)
                log_handlers.append(handler)
                worker_log_handlers.append(handler)
            fds = {
                OutputType.stdout: worker_run.stdout(),
                OutputType.stderr: worker_run.stderr(),
            }
            for output_type, fd in fds.items():
                selector.register(
                    fd,
                    selectors.EVENT_READ,
                    data=SelectorMeta(log_handlers=worker_log_handlers, output_type=output_type),
                )

        sidecar_runs_meta: dict[SidecarIndex, SidecarRunMeta] = {}
        for sidecar_index, sidecar_run in sidecar_runs.items():
            sidecar_log_handlers = []
            for factory in log_handler_factories:
                handler = factory.create_for_sidecar(sidecar_index, sidecar_run)
                log_handlers.append(handler)
                sidecar_log_handlers.append(handler)
            fds = {
                OutputType.stdout: sidecar_run.stdout(),
                OutputType.stderr: sidecar_run.stderr(),
            }
            for output_type, fd in fds.items():
                selector.register(
                    fd,
                    selectors.EVENT_READ,
                    data=SelectorMeta(log_handlers=sidecar_log_handlers, output_type=output_type),
                )
            sidecar_runs_meta[sidecar_index] = SidecarRunMeta(
                sidecar_run=sidecar_run,
                log_handlers=sidecar_log_handlers,
            )
        return ProcessManager(
            sidecar_runs=sidecar_runs_meta,
            worker_runs=worker_runs,
            io_selector=selector,
            log_handlers=log_handlers,
        )

    def _restart_sidecar(self, sidecar_index: SidecarIndex) -> None:
        sidecar_run_meta = self._sidecar_runs[sidecar_index]
        sidecar_run = sidecar_run_meta.sidecar_run
        log_handlers = sidecar_run_meta.log_handlers
        for fd in [sidecar_run.stdout(), sidecar_run.stderr()]:
            self._io_selector.unregister(fd)

        sidecar_run = sidecar_run.restart()
        new_sidecar_run_meta = SidecarRunMeta(
            sidecar_run=sidecar_run,
            log_handlers=log_handlers,
        )
        self._sidecar_runs[sidecar_index] = new_sidecar_run_meta

        fds = {
            OutputType.stdout: sidecar_run.stdout(),
            OutputType.stderr: sidecar_run.stderr(),
        }
        for output_type, fd in fds.items():
            self._io_selector.register(
                fd,
                selectors.EVENT_READ,
                data=SelectorMeta(log_handlers=log_handlers, output_type=output_type),
            )

    def poll(self) -> ProcessManagerPollStatus:
        status = self._poll_processes()
        self._process_logs()
        return status

    def _poll_processes(self) -> ProcessManagerPollStatus:
        exit_codes = [worker_run.poll() for worker_run in self._worker_runs.values()]
        match check_status(exit_codes):
            case PoolStatus.failed:
                return ProcessManagerPollStatus.fail
            case PoolStatus.success:
                return ProcessManagerPollStatus.success

        for sidecar_index, sidecar_run_meta in self._sidecar_runs.items():
            sidecar_run = sidecar_run_meta.sidecar_run
            match sidecar_run.need_restart():
                case RestartVerdict.restart:
                    print(f"Restart sidecar {sidecar_run.command}", file=sys.stderr)
                    self._restart_sidecar(sidecar_index)
                case RestartVerdict.fail:
                    print(f"Sidecar {sidecar_run.command} has been failed", file=sys.stderr)
                    return ProcessManagerPollStatus.fail
                case RestartVerdict.skip:
                    pass
                case RestartVerdict.unknown:
                    print(f"Warning: unknown restart policy for {sidecar_run.command}", file=sys.stderr)
                case _:
                    print(f"Warning: unknown restart verdict for {sidecar_run.command}", file=sys.stderr)
        return ProcessManagerPollStatus.running

    def _process_logs(self) -> None:
        for key, _ in self._io_selector.select(timeout=-1):
            if TYPE_CHECKING:
                assert isinstance(key, selectors.SelectorKey)
            meta: SelectorMeta = key.data
            lines = key.fileobj.readlines()  # type: ignore
            if not lines:
                continue
            for line in lines:
                line = line.rstrip()
                print(line, file=sys.stderr)
                record = LogRecord(
                    message=line,
                    fd=meta.output_type.value,
                    datetime=datetime.datetime.now(),
                )
                for handler in meta.log_handlers:
                    handler.process(record)

    def _stop(self) -> None:
        for worker_run in self._worker_runs.values():
            worker_run.terminate()
        for sidecar_run_meta in self._sidecar_runs.values():
            sidecar_run_meta.sidecar_run.terminate()
        self._io_selector.close()
        for handler in self._log_handlers:
            handler.stop()


def has_failed(exit_codes: list[Optional[int]]) -> bool:
    return any(code is not None and code != 0 for code in exit_codes)


def is_success(exit_codes: list[Optional[int]]) -> bool:
    return all(code == 0 for code in exit_codes)


class PoolStatus(enum.IntEnum):
    running = enum.auto()
    success = enum.auto()
    failed = enum.auto()


def check_status(exit_codes: list[Optional[int]]) -> PoolStatus:
    if has_failed(exit_codes):
        return PoolStatus.failed
    if is_success(exit_codes):
        return PoolStatus.success
    return PoolStatus.running



import attrs

from tractorun.sidecar import RestartPolicy


__all__ = ["TensorproxySidecar"]


@attrs.define(kw_only=True, slots=True, auto_attribs=True)
class TensorproxySidecar:
    enabled: bool
    restart_policy: RestartPolicy
    yt_path: str



import io

import torch


__all__ = ["TensorSerializer"]


class TensorSerializer:
    def serialize(self, tensor: object) -> bytes:
        buffer = io.BytesIO()
        torch.save(tensor, buffer)
        return buffer.getvalue()

    def desirialize(self, tensor: bytes, device: torch.device | None = None) -> dict:
        if device is None:
            device = torch.device("cpu")
        buffer = io.BytesIO(tensor)
        return torch.load(buffer, map_location=device)



import argparse
import os
import sys
from typing import Any
import uuid

import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.optim as optim
import torch.utils.data
import wandb
import yt.wrapper as yt

from tractorun.backend.tractorch import Tractorch
from tractorun.backend.tractorch.dataset import YtTensorDataset
from tractorun.backend.tractorch.serializer import TensorSerializer
from tractorun.env import EnvVariable
from tractorun.mesh import Mesh
from tractorun.resources import Resources
from tractorun.run import run
from tractorun.toolbox import Toolbox


DEFAULT_DATASET_PATH = "//home/yt-team/chiffa/tractorun/mnist/datasets/train"


class Net(nn.Module):
    def __init__(self) -> None:
        super(Net, self).__init__()
        self.l1 = torch.nn.Linear(28 * 28, 10)

    def forward(self, x: Any) -> torch.Tensor:
        return torch.relu(self.l1(x.view(x.size(0), -1)))


def train(toolbox: Toolbox) -> None:
    user_config = toolbox.get_user_config()
    dataset_path = user_config["dataset_path"]
    wandb_enabled = user_config["wandb_enabled"]
    if wandb_enabled:
        wandb.login(key=os.environ["WANDB_TOKEN"])
        wandb.init(
            project="tractorun",
            name="torch_mnist_checkpoints",
            id=user_config["wandb_run_id"],
        )

    device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
    serializer = TensorSerializer()
    print("Running on device:", device, file=sys.stderr)

    model = Net().to(device)
    optimizer = optim.Adadelta(model.parameters(), lr=1.0)

    checkpoint = toolbox.checkpoint_manager.get_last_checkpoint()
    first_batch_index = 0
    if checkpoint is not None:
        first_batch_index = checkpoint.metadata["first_batch_index"]
        print(
            "Found checkpoint with index",
            checkpoint.index,
            "and first batch index",
            first_batch_index,
            file=sys.stderr,
        )

        checkpoint_dict = serializer.desirialize(checkpoint.value)
        model.load_state_dict(checkpoint_dict["model"])
        optimizer.load_state_dict(checkpoint_dict["optimizer"])

    train_dataset = YtTensorDataset(
        toolbox,
        path=dataset_path,
        start=0,
        end=4000,
    )
    dataset_len = len(train_dataset)
    train_loader = torch.utils.data.DataLoader(train_dataset, batch_size=64)

    model.train()
    if wandb_enabled:
        wandb.watch(model, log_freq=100)

    final_loss = None

    for batch_idx, (data, target) in enumerate(train_loader):
        # TODO: Do it normally. YT dataloader? YT loader!
        if batch_idx < first_batch_index:
            continue

        if batch_idx == 10 and first_batch_index == 0:
            assert False, "Force restart to test checkpoints"

        data, target = data.to(device), target.to(device)
        optimizer.zero_grad()
        output = model(data)
        loss = F.nll_loss(output, target)
        loss.backward()
        optimizer.step()
        if batch_idx % 5 == 0:
            print(
                "Train[{}/{} ({:.0f}%)]\tLoss: {:.6f}".format(
                    batch_idx * len(data),
                    dataset_len,
                    100.0 * batch_idx / len(train_loader),
                    loss.item(),
                ),
                file=sys.stderr,
            )
            if wandb_enabled:
                wandb.log({"loss": loss.item(), "batch_idx": batch_idx})

        if batch_idx % 3 == 0:
            state_dict = {
                "model": model.state_dict(),
                "optimizer": optimizer.state_dict(),
            }
            metadata_dict = {
                "first_batch_index": batch_idx + 1,
                "loss": loss.item(),
            }
            toolbox.checkpoint_manager.save_checkpoint(serializer.serialize(state_dict), metadata_dict)
            # save checkpoint synchronously
            # task = toolbox.checkpoint_manager.save_checkpoint(serializer.serialize(state_dict), metadata_dict)
            # task.wait(timeout=10)
            print("Saved checkpoint after batch with index", batch_idx, file=sys.stderr)
        final_loss = loss.item()

    # Save the model
    model_path = toolbox.save_model(
        data=serializer.serialize(model.state_dict()),
        dataset_path=dataset_path,
        metadata={
            "loss": str(final_loss),
        },
    )
    print("Model saved to", model_path, file=sys.stderr)


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("--yt-home-dir", type=str, required=True)
    parser.add_argument("--dataset-path", type=str, default=DEFAULT_DATASET_PATH)
    parser.add_argument("--wandb", action=argparse.BooleanOptionalAction, default=False)  # type: ignore  # FIXME: min python -> 3.10
    args = parser.parse_args()

    # Remove old checkpoints.
    workdir = args.yt_home_dir
    if yt.exists(f"{workdir}/checkpoints"):
        yt.remove(f"{workdir}/checkpoints", recursive=True)

    mesh = Mesh(node_count=1, process_per_node=1, gpu_per_process=0)
    run(
        train,
        backend=Tractorch(),
        yt_path=workdir,
        mesh=mesh,
        user_config={
            "workdir": workdir,
            "dataset_path": args.dataset_path,
            "wandb_run_id": str(uuid.uuid4()),
            "wandb_enabled": args.wandb,
        },
        env=[
            EnvVariable(
                name="WANDB_API_KEY",
                cypress_path=os.environ.get("WANDB_SECRET"),
            ),
        ],
        resources=Resources(memory_limit=4 * (1024**3)),
    )



import sys

from jax import (
    grad,
    jit,
)

from tractorun.backend.tractorax import Tractorax
from tractorun.mesh import Mesh
from tractorun.resources import Resources
from tractorun.run import run
from tractorun.toolbox import Toolbox


def task(_: Toolbox) -> None:
    @jit
    def f(x: int) -> int:
        return x**2 + 3 * x + 1

    grad_f = grad(f)
    print(grad_f(1.0), file=sys.stderr)


if __name__ == "__main__":
    mesh = Mesh(node_count=1, process_per_node=1, gpu_per_process=0, pool_trees=["gpu_h100"])
    run(
        task,
        backend=Tractorax(),
        yt_path="//home/gritukan/mnist/trainings/dense",
        mesh=mesh,
        docker_image="cr.ai.nebius.cloud/crnf2coti090683j5ssi/tractorun/tractorax_runtime:2024-06-18-16-13-57",
        resources=Resources(cpu_limit=4.0, memory_limit=16 * 1024**3),
        local=True,
    )



from typing import Any

import attrs
from yt import wrapper as yt

from tractorun.docker_auth import (
    DockerAuthData,
    DockerAuthSecret,
)
from tractorun.exception import DockerAuthDataError


@attrs.define(kw_only=True, slots=True, auto_attribs=True)
class DockerAuthInternal:
    username: str | None
    password: str | None
    auth: str | None

    def to_spec(self) -> dict:
        data = {}
        if self.username:
            data["username"] = self.username
        if self.password:
            data["password"] = self.password
        if self.auth:
            data["auth"] = self.auth
        return data


@attrs.define(kw_only=True, slots=True, auto_attribs=True)
class DockerAuthDataExtractor:
    _yt_client: yt.YtClient

    def _is_raw_auth_data_valid(self, raw_auth_data: Any) -> bool:
        if not isinstance(raw_auth_data, dict):
            return False
        if set(raw_auth_data.keys()) != {"username", "password"} and set(raw_auth_data.keys()) != {"auth"}:
            return False
        if not all([isinstance(value, str) for value in raw_auth_data.values()]):
            return False
        return True

    def extract(self, auth_data: DockerAuthData) -> DockerAuthInternal:
        match auth_data:
            case DockerAuthSecret():
                assert isinstance(auth_data, DockerAuthSecret)  # skip type warning in pycharm
                raw_auth_data = self._yt_client.get(auth_data.cypress_path)
                if not self._is_raw_auth_data_valid(raw_auth_data):
                    format_options = '{username="..."; password="...";} or {auth="...";}'
                    raise DockerAuthDataError(f"docker_auth document should have format {format_options}")
                return DockerAuthInternal(
                    username=raw_auth_data.get("username"),
                    password=raw_auth_data.get("password"),
                    auth=raw_auth_data.get("auth"),
                )
            case _:
                raise DockerAuthDataError("Unknown auth data type")



YT_USER_CONFIG_ENV_VAR: str = "YT_USER_CONFIG"
TRACTO_CONFIG_ENV_VAR: str = "TRACTO_CONFIG"
BIND_PATHS_ENV_VAR: str = "BIND_PATHS"
BOOTSTRAP_CONFIG_FILENAME_ENV_VAR: str = "BOOTSTRAP_CONFIG_FILENAME"
DEFAULT_CLUSTER_CONFIG_PATH: str = "//home/tractorun/config"
DEFAULT_OPERATION_TITLE: str = "Tractorun operation"

DEFAULT_DOCKER_IMAGE: str = "cr.ai.nebius.cloud/crnf2coti090683j5ssi/tractorun/torchesaurus_runtime:2024-09-21-01-01-28"
BOOTSTRAP_CONFIG_NAME: str = "__bootstrap_config"

USER_DESCRIPTION_MANAGER_NAME = "extra"
TRACTORUN_DESCRIPTION_MANAGER_NAME = "tractorun"



import uuid

import attrs
import yt.wrapper as yt


@attrs.define(kw_only=True, slots=True, auto_attribs=True)
class TrainingDir:
    base_path: str
    incarnations_path: str
    models_path: str
    checkpoints_path: str
    primary_lock_path: str
    logs_path: str
    worker_logs_path: str
    sidecar_logs_path: str

    @classmethod
    def create(cls, path: str) -> "TrainingDir":
        return TrainingDir(
            base_path=path,
            primary_lock_path=path + "/primary_lock",
            incarnations_path=path + "/incarnations",
            models_path=path + "/models",
            checkpoints_path=path + "/checkpoints",
            logs_path=path + "/logs",
            worker_logs_path=path + f"/_raw_logs/{uuid.uuid4()}",
            sidecar_logs_path=path + f"/_raw_logs/{uuid.uuid4()}",
        )

    def get_incarnation_path(self, incarnation_id: int) -> str:
        return self.incarnations_path + f"/{incarnation_id:05d}"


def prepare_training_dir(training_dir: TrainingDir, yt_client: yt.YtClient) -> None:
    yt_client.create(
        "map_node", training_dir.base_path, attributes={"incarnation_id": -1}, recursive=True, ignore_existing=True
    )
    yt_client.create("map_node", training_dir.primary_lock_path, ignore_existing=True)
    yt_client.create("map_node", training_dir.incarnations_path, ignore_existing=True)
    yt_client.create("map_node", training_dir.models_path, ignore_existing=True)
    yt_client.create("map_node", training_dir.checkpoints_path, ignore_existing=True)
    yt_client.create("map_node", training_dir.logs_path, ignore_existing=True)
    yt_client.create("map_node", training_dir.worker_logs_path, recursive=True, ignore_existing=True)
    yt_client.create("map_node", training_dir.sidecar_logs_path, recursive=True, ignore_existing=True)



import uuid

from tests.utils import (
    DOCKER_IMAGE,
    make_cli_args,
    make_run_config,
    run_config_file,
)
from tractorun.backend.generic import GenericBackend
from tractorun.cli.tractorun_runner import make_configuration
from tractorun.mesh import Mesh
from tractorun.run import run
from tractorun.toolbox import Toolbox


def test_configuration_pool() -> None:
    _, _, config = make_configuration(make_cli_args("--mesh.pool", "cli"))
    assert config.mesh.pool == "cli"

    run_config = make_run_config({"mesh": {"pool": "config"}})
    with run_config_file(run_config) as run_config_path:
        _, _, config = make_configuration(["--run-config-path", run_config_path])
    assert config.mesh.pool == "config"

    with run_config_file(run_config) as run_config_path:
        _, _, config = make_configuration(["--run-config-path", run_config_path, "--mesh.pool", "cli"])
    assert config.mesh.pool == "cli"


def test_configuration_pool_trees() -> None:
    _, _, config = make_configuration(make_cli_args("--mesh.pool-trees", "cli1", "--mesh.pool-trees", "cli2"))
    assert config.mesh.pool_trees == ["cli1", "cli2"]

    run_config = make_run_config({"mesh": {"pool_trees": ["config1", "config2"]}})
    with run_config_file(run_config) as run_config_path:
        _, _, config = make_configuration(["--run-config-path", run_config_path])
    assert config.mesh.pool_trees == ["config1", "config2"]

    with run_config_file(run_config) as run_config_path:
        _, _, config = make_configuration(
            make_cli_args(
                "--run-config-path", run_config_path, "--mesh.pool-trees", "cli1", "--mesh.pool-trees", "cli2"
            )
        )
    assert config.mesh.pool_trees == ["cli1", "cli2"]


def test_run_pickle(yt_path: str) -> None:
    def checker(toolbox: Toolbox) -> None:
        pass

    title = f"super title {uuid.uuid4()}"

    run_info = run(
        checker,
        yt_path=yt_path,
        title=title,
        mesh=Mesh(node_count=1, process_per_node=1, gpu_per_process=0, pool_trees=["some_tree"], pool="some_pool"),
        backend=GenericBackend(),
        docker_image=DOCKER_IMAGE,
        dry_run=True,
    )
    assert run_info.operation_spec["pool"] == "some_pool"
    assert run_info.operation_spec["pool_trees"] == ["some_tree"]



import sys

import jax

from tractorun.backend.tractorax import Tractorax
from tractorun.mesh import Mesh
from tractorun.resources import Resources
from tractorun.run import run
from tractorun.toolbox import Toolbox


def task(_: Toolbox) -> None:
    print("Total device count:", jax.device_count(), file=sys.stderr)
    print("Local device count:", jax.local_device_count(), file=sys.stderr)

    xs = jax.numpy.ones(jax.local_device_count())
    print(jax.pmap(lambda x: jax.lax.psum(x, "i"), axis_name="i")(xs), file=sys.stderr)


if __name__ == "__main__":
    mesh = Mesh(node_count=1, process_per_node=7, gpu_per_process=1, pool_trees=["gpu-h100"])
    run(
        task,
        backend=Tractorax(),
        yt_path="//home/gritukan/mnist/trainings/dense",
        mesh=mesh,
        docker_image="cr.ai.nebius.cloud/crnf2coti090683j5ssi/tractorun/tractorax_runtime:2024-06-18-17-42-54",
        resources=Resources(cpu_limit=4.0, memory_limit=64 * 1024**3),
    )



import torch
import torch.distributed as dist

from tractorun.backend.tractorch import Tractorch
from tractorun.mesh import Mesh
from tractorun.run import run
from tractorun.toolbox import Toolbox


def train(toolbox: Toolbox) -> None:
    device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
    # device = torch.device('cpu')
    print("Running on device:", device)

    if toolbox.coordinator.get_self_index() == 0:
        x = torch.as_tensor([2.0], device=device)
        y = torch.as_tensor([3.0], device=device)
        print(f"Sending summands to peer 1 (X = {x}, Y = {y})")
        dist.send(x, dst=1)
        dist.send(y, dst=1)
        print("Waiting for result from peer 1")
        sum = torch.zeros(1, device=device)
        dist.recv(tensor=sum, src=1)
        print(f"Received sum from peer 1: {sum}")
    else:
        x = torch.zeros(1, device=device)
        print("Waiting for X from peer 0")
        dist.recv(tensor=x, src=0)
        print(f"Received X from peer 0: {x}")
        print("Waiting for Y from peer 0")
        y = torch.zeros(1, device=device)
        dist.recv(tensor=y, src=0)
        print(f"Received Y from peer 1: {y}")
        sum = x + y
        print(f"Sending sum to peer 1: {sum}")
        dist.send(tensor=sum, dst=0)


if __name__ == "__main__":
    mesh = Mesh(1, 2, 0)
    run(
        train,
        backend=Tractorch(),
        yt_path="//home/gritukan/train",
        mesh=mesh,
    )



import abc

from tractorun.private.closet import Closet as _Closet


__all__ = ["EnvironmentBase", "BackendBase"]


class EnvironmentBase(abc.ABC):
    @abc.abstractmethod
    def prepare(self, closet: _Closet) -> None:
        pass


class BackendBase(abc.ABC):
    @property
    @abc.abstractmethod
    def environment(self) -> EnvironmentBase:
        pass



import argparse
import os
import sys
from typing import (
    Any,
    Tuple,
)
import uuid

from pytorch_lightning import (
    LightningModule,
    Trainer,
)
from pytorch_lightning.loggers import WandbLogger
from pytorch_lightning.utilities.types import STEP_OUTPUT
import torch
from torch import Tensor
from torch.nn import functional as F
from torch.utils.data import DataLoader
import wandb

from tractorun.backend.tractorch import Tractorch
from tractorun.backend.tractorch.dataset import YtTensorDataset
from tractorun.env import EnvVariable
from tractorun.mesh import Mesh
from tractorun.resources import Resources
from tractorun.run import run
from tractorun.toolbox import Toolbox


DEFAULT_DATASET_PATH = "//home/yt-team/chiffa/tractorun/mnist/datasets/train"


class MNISTModel(LightningModule):
    def __init__(self) -> None:
        super().__init__()
        self.l1 = torch.nn.Linear(28 * 28, 50)
        self.l2 = torch.nn.Linear(50, 10)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = torch.relu(self.l1(x.view(x.size(0), -1)))
        x = torch.relu(self.l2(x))
        return x

    def training_step(self, batch: Tuple[Tensor, ...], batch_nb: Any) -> STEP_OUTPUT:
        x, y = batch
        loss = F.cross_entropy(self(x), y)
        self.log("train_loss", loss)
        self.log("batch_nb", batch_nb)
        return loss

    def configure_optimizers(self) -> torch.optim.Adam:
        return torch.optim.Adam(self.parameters(), lr=0.02)


def train(toolbox: Toolbox) -> None:
    user_config = toolbox.get_user_config()
    dataset_path = user_config["dataset_path"]
    wandb_enabled = user_config["wandb_enabled"]
    if wandb_enabled:
        wandb.login(key=os.environ["WANDB_TOKEN"])
        wandb_logger = WandbLogger(
            project="tractorun",
            name="lightning_mnist_ddp",
            id=user_config["wandb_run_id"],
            log_model="all",
        )
    else:
        wandb_logger = None

    print("Mesh", toolbox.mesh, file=sys.stderr)
    device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
    print("Running on device:", device, file=sys.stderr)

    mnist_model = MNISTModel()
    train_dataset = YtTensorDataset(toolbox, dataset_path)
    train_loader = DataLoader(train_dataset, batch_size=64)

    trainer = Trainer(
        max_epochs=3,
        devices=toolbox.mesh.process_per_node,
        num_nodes=toolbox.mesh.node_count,
        strategy="ddp",
        logger=wandb_logger,
    )

    trainer.fit(mnist_model, train_loader)


def main() -> None:
    parser = argparse.ArgumentParser()
    parser.add_argument("--yt-home-dir", type=str, required=True)
    parser.add_argument("--dataset-path", type=str, default=DEFAULT_DATASET_PATH)
    parser.add_argument("--wandb", action=argparse.BooleanOptionalAction, default=False)  # type: ignore  # FIXME: min python -> 3.10
    args = parser.parse_args()

    workdir = args.yt_home_dir
    mesh = Mesh(node_count=1, process_per_node=3, gpu_per_process=1)
    run(
        train,
        backend=Tractorch(),
        yt_path=workdir,
        mesh=mesh,
        resources=Resources(
            memory_limit=8076021002,
        ),
        env=[
            EnvVariable(
                name="WANDB_API_KEY",
                cypress_path=os.environ.get("WANDB_SECRET"),
            ),
        ],
        user_config={
            "dataset_path": args.dataset_path,
            "wandb_run_id": str(uuid.uuid4()),
            "wandb_enabled": args.wandb,
        },
    )


if __name__ == "__main__":
    main()



import json
import sys
import threading

from _pytest.capture import CaptureFixture
import attrs
import pytest
from yt.wrapper import YtOperationFailedError

from tests.utils import (
    DOCKER_IMAGE,
    TractoCli,
    get_data_path,
    make_cli_args,
    make_run_config,
    run_config_file,
)
from tests.yt_instances import YtInstance
from tractorun.backend.generic import GenericBackend
from tractorun.cli.tractorun_runner import make_configuration
from tractorun.mesh import Mesh
from tractorun.operation_log import OperationLogMode
from tractorun.private.stderr_reader import (
    STDERR_READER_THREAD_NAME,
    YtStderrReader,
)
from tractorun.run import run
from tractorun.stderr_reader import StderrMode
from tractorun.toolbox import Toolbox


TEST_STRINGS = ["hello", "my dear", "friend"]


def test_configuration() -> None:
    _, _, config = make_configuration(make_cli_args())
    assert config.proxy_stderr_mode == StderrMode.disabled

    _, _, config = make_configuration(make_cli_args("--proxy-stderr-mode", StderrMode.primary.value))
    assert config.proxy_stderr_mode == StderrMode.primary

    run_config = make_run_config({"proxy_stderr_mode": StderrMode.primary.value})
    with run_config_file(run_config) as run_config_path:
        _, _, config = make_configuration(["--run-config-path", run_config_path])
    assert config.proxy_stderr_mode == StderrMode.primary

    with run_config_file(run_config) as run_config_path:
        _, _, config = make_configuration(
            ["--run-config-path", run_config_path, "--proxy-stderr-mode", StderrMode.disabled],
        )
    assert config.proxy_stderr_mode == StderrMode.disabled


@attrs.define(kw_only=True, slots=True, auto_attribs=True)
class MockStderrProvider:
    _lines: list[bytes]
    _index: int = attrs.field(default=0)

    def read(self) -> bytes | None:
        while self._index < len(self._lines):
            line = self._lines[self._index]
            self._index += 1
            return line
        return None


@pytest.mark.parametrize(
    "lines,expected",
    [
        (
            [b"a", b"ab"],
            [b"a", b"b"],
        ),
        (
            [b"abcd", b"abcde"],
            [b"abcd", b"e"],
        ),
        (
            [b"abcd", b"abcdef"],
            [b"abcd", b"ef"],
        ),
        (
            [b"", b"a"],
            [b"", b"a"],
        ),
        (
            [b"a", b"b"],
            [b"a", b"b"],
        ),
        (
            [b"ab", b"ab"],
            [b"ab", b""],
        ),
        (
            [b"a", b"abcde", b"defg", b"abc"],
            [b"a", b"bcde", b"fg", b"abc"],
        ),
    ],
)
def test_stderr_reader(lines: list[bytes], expected: list[bytes]) -> None:
    mock = MockStderrProvider(lines=lines)
    reader = YtStderrReader(stderr_getter=mock.read, stop_on_none=True)
    result = list(reader.get_output())
    assert result == expected


@pytest.mark.parametrize(
    "mode",
    [StderrMode.primary, StderrMode.disabled],
)
def test_operation_pickling(
    mode: StderrMode, yt_path: str, yt_instance: YtInstance, capsys: CaptureFixture[str]
) -> None:
    def checker(toolbox: Toolbox) -> None:
        marker = "primary" if toolbox.coordinator.is_primary() else "secondary"
        for test_string in TEST_STRINGS:
            print(f"{marker} {test_string}", file=sys.stderr)

    yt_client = yt_instance.get_client()
    mesh = Mesh(node_count=2, process_per_node=1, gpu_per_process=0)
    run(
        checker,
        backend=GenericBackend(),
        yt_path=yt_path,
        mesh=mesh,
        yt_client=yt_client,
        docker_image=DOCKER_IMAGE,
        proxy_stderr_mode=mode,
    )
    captured = capsys.readouterr()
    match mode:
        case StderrMode.disabled:
            for s in TEST_STRINGS:
                assert f"{s}\n" not in captured.out
        case StderrMode.primary:
            for s in TEST_STRINGS:
                assert f"primary {s}\n" in captured.out
            for s in TEST_STRINGS:
                assert f"secondary {s}\n" not in captured.out
        case _:
            raise Exception(f"Unknown mode {mode}")


def test_with_realtime_table(yt_path: str, yt_instance: YtInstance, capsys: CaptureFixture[str]) -> None:
    def checker(toolbox: Toolbox) -> None:
        for test_string in TEST_STRINGS:
            print(test_string)

    yt_client = yt_instance.get_client()
    mesh = Mesh(node_count=1, process_per_node=1, gpu_per_process=0)
    run(
        checker,
        backend=GenericBackend(),
        yt_path=yt_path,
        mesh=mesh,
        yt_client=yt_client,
        docker_image=DOCKER_IMAGE,
        proxy_stderr_mode=StderrMode.primary,
        operation_log_mode=OperationLogMode.realtime_yt_table,
    )
    captured = capsys.readouterr()
    for s in TEST_STRINGS:
        assert f"{s}\n" in captured.out


def test_operation_cli(yt_instance: YtInstance, yt_path: str) -> None:
    yt_client = yt_instance.get_client()

    tracto_cli = TractoCli(
        command=["python3", "/tractorun_tests/stderr_script.py"],
        args=[
            "--yt-path",
            yt_path,
            "--user-config",
            json.dumps(
                {
                    "test_strings": TEST_STRINGS,
                },
            ),
            "--bind-local",
            f"{get_data_path('../data/stderr_script.py')}:/tractorun_tests/stderr_script.py",
            "--proxy-stderr-mode",
            StderrMode.primary,
        ],
    )
    op_run = tracto_cli.run()
    assert op_run.is_exitcode_valid()
    assert op_run.is_operation_state_valid(yt_client=yt_client, job_count=1)

    stdout = op_run.stdout.decode("utf-8")
    for s in TEST_STRINGS:
        assert f"{s}\n" in stdout


@pytest.mark.parametrize(
    "mode",
    [StderrMode.primary, StderrMode.disabled],
)
def test_stop_on_fail(mode: StderrMode, yt_path: str, yt_instance: YtInstance) -> None:
    def checker(toolbox: Toolbox) -> None:
        print("message", file=sys.stderr)
        raise Exception("fail operation")

    yt_client = yt_instance.get_client()
    mesh = Mesh(node_count=1, process_per_node=1, gpu_per_process=0)
    with pytest.raises(YtOperationFailedError):
        run(
            checker,
            backend=GenericBackend(),
            yt_path=yt_path,
            mesh=mesh,
            yt_client=yt_client,
            docker_image=DOCKER_IMAGE,
            proxy_stderr_mode=mode,
            yt_operation_spec={"max_failed_job_count": 1},
        )
    names = [thread.name for thread in threading.enumerate()]
    assert STDERR_READER_THREAD_NAME not in names


def test_multiple_processes(yt_path: str, yt_instance: YtInstance, capsys: CaptureFixture[str]) -> None:
    message = "test message"

    def checker(toolbox: Toolbox) -> None:
        print(message, file=sys.stderr)

    yt_client = yt_instance.get_client()
    mesh = Mesh(node_count=1, process_per_node=2, gpu_per_process=0)
    run(
        checker,
        backend=GenericBackend(),
        yt_path=yt_path,
        mesh=mesh,
        yt_client=yt_client,
        docker_image=DOCKER_IMAGE,
        proxy_stderr_mode=StderrMode.primary,
    )
    captured = capsys.readouterr()
    assert message in captured.out


def test_read_two_operations(yt_path: str, yt_instance: YtInstance, capsys: CaptureFixture[str]) -> None:
    key = "iter_id"

    def checker(toolbox: Toolbox) -> None:
        user_config = toolbox.get_user_config()
        id_ = user_config[key]
        print(f"message {id_}", file=sys.stderr)

    yt_client = yt_instance.get_client()
    mesh = Mesh(node_count=1, process_per_node=1, gpu_per_process=0)
    run(
        checker,
        backend=GenericBackend(),
        yt_path=yt_path,
        mesh=mesh,
        yt_client=yt_client,
        docker_image=DOCKER_IMAGE,
        proxy_stderr_mode=StderrMode.primary,
        user_config={key: "1"},
    )
    captured = capsys.readouterr()
    assert "message 1" in captured.out

    run(
        checker,
        backend=GenericBackend(),
        yt_path=yt_path,
        mesh=mesh,
        yt_client=yt_client,
        docker_image=DOCKER_IMAGE,
        proxy_stderr_mode=StderrMode.primary,
        user_config={key: "2"},
    )
    captured = capsys.readouterr()
    assert "message 2" in captured.out, captured.err



import sys
from typing import (
    Any,
    Tuple,
)

from pytorch_lightning import (
    LightningModule,
    Trainer,
)
from pytorch_lightning.utilities.types import STEP_OUTPUT
import torch
from torch import Tensor
from torch.nn import functional as F
from torch.utils.data import DataLoader

from tractorun.backend.tractorch import Tractorch
from tractorun.backend.tractorch.dataset import YtTensorDataset
from tractorun.mesh import Mesh
from tractorun.resources import Resources
from tractorun.run import run
from tractorun.toolbox import Toolbox


class MNISTModel(LightningModule):
    def __init__(self) -> None:
        super().__init__()
        self.l1 = torch.nn.Linear(28 * 28, 10)

    def forward(self, x: Any) -> torch.Tensor:
        return torch.relu(self.l1(x.view(x.size(0), -1)))

    def training_step(self, batch: Tuple[Tensor, ...], batch_nb: Any) -> STEP_OUTPUT:
        x, y = batch
        loss = F.cross_entropy(self(x), y)
        return loss

    def configure_optimizers(self) -> torch.optim.Adam:
        return torch.optim.Adam(self.parameters(), lr=0.02)


def train(toolbox: Toolbox) -> None:
    device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
    print("Running on device:", device, file=sys.stderr)

    mnist_model = MNISTModel()
    train_dataset = YtTensorDataset(
        toolbox,
        "//home/gritukan/mnist/datasets/train",
        start=0,
        end=1000,
    )
    train_loader = DataLoader(train_dataset, batch_size=64)

    trainer = Trainer(max_epochs=3)

    trainer.fit(mnist_model, train_loader)


if __name__ == "__main__":
    mesh = Mesh(node_count=1, process_per_node=1, gpu_per_process=0)
    run(
        train,
        backend=Tractorch(),
        yt_path="//home/gritukan/mnist/trainings/dense",
        mesh=mesh,
        resources=Resources(
            memory_limit=1076021002,
        ),
    )



import sys
from typing import (
    Any,
    Tuple,
)

from pytorch_lightning import (
    LightningModule,
    Trainer,
)
from pytorch_lightning.utilities.types import STEP_OUTPUT
import torch
from torch import Tensor
from torch.nn import functional as F
from torch.utils.data import DataLoader

from tractorun.backend.tractorch import (
    Tractorch,
    YtTensorDataset,
)
from tractorun.run import prepare_and_get_toolbox


module_locations = ["./tmpfs/modules", "./modules"]
sys.path = sys.path + module_locations


class MNISTModel(LightningModule):
    def __init__(self) -> None:
        super().__init__()
        self.l1 = torch.nn.Linear(28 * 28, 50)
        self.l2 = torch.nn.Linear(50, 10)

    def forward(self, x: Any) -> torch.Tensor:
        x = torch.relu(self.l1(x.view(x.size(0), -1)))
        x = torch.relu(self.l2(x))
        return x

    def training_step(self, batch: Tuple[Tensor, ...], batch_nb: Any) -> STEP_OUTPUT:
        x, y = batch
        loss = F.cross_entropy(self(x), y)
        return loss

    def configure_optimizers(self) -> torch.optim.Adam:
        return torch.optim.Adam(self.parameters(), lr=0.02)


toolbox = prepare_and_get_toolbox(backend=Tractorch())
device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
print("Running on device:", device, file=sys.stderr)

mnist_model = MNISTModel()
train_dataset = YtTensorDataset(toolbox, "//home/gritukan/mnist/datasets/train")
train_loader = DataLoader(train_dataset, batch_size=64)

trainer = Trainer(
    max_epochs=3,
    devices=toolbox.mesh.process_per_node,
    num_nodes=toolbox.mesh.node_count,
    strategy="ddp",
)

trainer.fit(mnist_model, train_loader)



__all__ = [
    "BaseTractorunException",
    "StderrReaderError",
    "TractorunConfigError",
    "TractorunWandbError",
    "TractorunConfigurationError",
    "DockerAuthDataError",
    "TractorunVersionMismatchError",
    "TractorunBootstrapError",
]


class BaseTractorunException(Exception):
    pass


class StderrReaderError(BaseTractorunException):
    pass


class TractorunBootstrapError(BaseTractorunException):
    pass


class TractorunConfigError(BaseTractorunException):
    pass


class TractorunVersionMismatchError(BaseTractorunException):
    pass


class TractorunConfigurationError(BaseTractorunException):
    pass


class TractorunWandbError(BaseTractorunException):
    pass


class DockerAuthDataError(BaseTractorunException):
    pass



from typing import Optional

import pytest

from tractorun.private.process_manager import (
    PoolStatus,
    check_status,
    has_failed,
    is_success,
)


@pytest.mark.parametrize(
    "exit_codes,expected",
    [
        ([1], True),
        ([0], False),
        ([None], False),
        ([None, 0], False),
        ([None, 1], True),
        ([None, 0, 1], True),
        ([0, 0], False),
        ([1, 0], True),
        ([1, 1], True),
    ],
)
def test_has_failed(exit_codes: list[Optional[int]], expected: bool) -> None:
    assert has_failed(exit_codes) == expected


@pytest.mark.parametrize(
    "exit_codes,expected",
    [
        ([1], False),
        ([0], True),
        ([None], False),
        ([None, 0], False),
        ([None, 1], False),
        ([None, 0, 1], False),
        ([0, 0], True),
        ([1, 0], False),
        ([1, 1], False),
    ],
)
def test_is_success(exit_codes: list[Optional[int]], expected: bool) -> None:
    assert is_success(exit_codes) == expected


@pytest.mark.parametrize(
    "exit_codes,expected",
    [
        ([1], PoolStatus.failed),
        ([0], PoolStatus.success),
        ([None], PoolStatus.running),
        ([None, None], PoolStatus.running),
        ([None, 0], PoolStatus.running),
        ([None, 1], PoolStatus.failed),
        ([None, 0, 1], PoolStatus.failed),
        ([0, 0], PoolStatus.success),
        ([1, 0], PoolStatus.failed),
        ([1, 1], PoolStatus.failed),
    ],
)
def test_check_status(exit_codes: list[Optional[int]], expected: bool) -> None:
    assert check_status(exit_codes) == expected



import attrs


__all__ = ["Mesh"]


@attrs.define
class Mesh:
    node_count: int = attrs.field()
    process_per_node: int = attrs.field()
    gpu_per_process: int = attrs.field()
    pool: str | None = attrs.field(default=None)
    pool_trees: list[str] | None = attrs.field(default=None)

    @property
    def peer_count(self) -> int:
        return self.node_count * self.process_per_node



import subprocess
import sys

from tractorun.backend.generic import GenericBackend
from tractorun.run import prepare_and_get_toolbox


def run_cmd(cmd: list) -> None:
    result = subprocess.run(cmd, capture_output=True, text=True)
    print(result.stdout, file=sys.stderr)
    print(result.stderr, file=sys.stderr)
    assert result.returncode == 0, f"command failed with return code {result.returncode}"


toolbox = prepare_and_get_toolbox(backend=GenericBackend())

print("Running ibstatus", file=sys.stderr)
run_cmd(["ibstatus"])

if toolbox.coordinator.is_primary():
    print("Running ib_send_bw as primary", file=sys.stderr)
    run_cmd(["ib_send_bw", "--report_gbits", "-D", "10"])
else:
    print("Running ib_send_bw as secondary", file=sys.stderr)
    primary_address = toolbox.coordinator.get_primary_endpoint()
    print("Primary address", primary_address)
    # primary_host, primary_port = toolbox.coordinator.get_primary_endpoint().split(":")

    # primary_ip = socket.getaddrinfo(primary_host, primary_port)[0][4][0]
    if "e00ypsy4gwrt0b2920" in primary_address:
        primary_ip = "192.168.0.36"
    else:
        assert "e00q0mwgg70qbs7fzd" in primary_address
        primary_ip = "192.168.0.27"

    print("Primary endpoint is ", primary_ip, file=sys.stderr)
    cmd = ["ib_send_bw", "--report_gbits", "-D", "10", primary_ip]
    print(f"run command {cmd}")
    run_cmd(cmd)



import json
import os
import pathlib
import typing as t

import pytest

from tests.utils import (
    DOCKER_IMAGE,
    TractoCli,
    get_data_path,
    make_cli_args,
    make_run_config,
    run_config_file,
)
from tests.yt_instances import YtInstance
from tractorun.backend.generic import GenericBackend
from tractorun.bind import (
    BindCypress,
    BindLocal,
)
from tractorun.cli.tractorun_runner import make_configuration
from tractorun.mesh import Mesh
from tractorun.run import run
from tractorun.toolbox import Toolbox


FILE_PATH = "/bind/file"
DIR_PATH = "/bind/dir"


@pytest.fixture
def cypress_file(yt_instance: YtInstance, yt_path: str) -> str:
    yt_client = yt_instance.get_client()
    file_path = f"{yt_path}/cypress_file"
    yt_client.write_file(file_path, b"hello")
    yt_client.set_attribute(file_path, "executable", True)
    return file_path


def test_configuration_cypress() -> None:
    _, _, config = make_configuration(
        make_cli_args("--bind-cypress", "//tmp/cli:cli", "command"),
    )
    assert config.bind_cypress == [BindCypress(source="//tmp/cli", destination="cli")]

    run_config = make_run_config(
        {
            "bind_cypress": ["//tmp/config:config"],
        }
    )
    with run_config_file(run_config) as run_config_path:
        _, _, config = make_configuration(["--run-config-path", run_config_path])
    assert config.bind_cypress == [BindCypress(source="//tmp/config", destination="config")]

    with run_config_file(run_config) as run_config_path:
        _, _, config = make_configuration(
            make_cli_args("--run-config-path", run_config_path, "--bind-cypress", "//tmp/cli:cli"),
        )
    assert config.bind_cypress == [BindCypress(source="//tmp/cli", destination="cli")]


def test_configuration_local() -> None:
    _, _, config = make_configuration(make_cli_args("--bind-local", "//tmp/cli:/tmp/cli"))
    assert config.bind_local == [BindLocal(source="//tmp/cli", destination="/tmp/cli")]

    run_config = make_run_config(
        {
            "bind_local": ["//tmp/config:/tmp/config"],
        }
    )
    with run_config_file(run_config) as run_config_path:
        _, _, config = make_configuration(["--run-config-path", run_config_path])
    assert config.bind_local == [BindLocal(source="//tmp/config", destination="/tmp/config")]

    with run_config_file(run_config) as run_config_path:
        _, _, config = make_configuration(["--run-config-path", run_config_path, "--bind-local", "//tmp/cli:/tmp/cli"])
    assert config.bind_local == [BindLocal(source="//tmp/cli", destination="/tmp/cli")]


def checker_dir(toolbox: Toolbox) -> None:
    assert pathlib.Path(DIR_PATH).is_dir()
    assert (pathlib.Path(DIR_PATH) / "some_file").is_file()


def get_file_checker(file_path: str = FILE_PATH, check_executable: bool = False) -> t.Callable[[Toolbox], None]:
    def checker_file(toolbox: Toolbox) -> None:
        assert pathlib.Path(file_path).is_file()
        if check_executable:
            assert os.access(file_path, os.X_OK)

    return checker_file


def test_local_bind_file_pickle(yt_instance: YtInstance, yt_path: str) -> None:
    yt_client = yt_instance.get_client()

    run(
        get_file_checker(),
        backend=GenericBackend(),
        yt_path=yt_path,
        binds_local=[
            BindLocal(
                source=get_data_path("../data/binds/another_file"),
                destination=FILE_PATH,
            ),
        ],
        mesh=Mesh(node_count=1, process_per_node=1, gpu_per_process=0),
        yt_client=yt_client,
        docker_image=DOCKER_IMAGE,
    )


def test_local_bind_dir_pickle(yt_instance: YtInstance, yt_path: str) -> None:
    yt_client = yt_instance.get_client()

    run(
        checker_dir,
        backend=GenericBackend(),
        yt_path=yt_path,
        binds_local=[
            BindLocal(
                source=get_data_path("../data/binds/bind_dir"),
                destination=DIR_PATH,
            ),
        ],
        mesh=Mesh(node_count=1, process_per_node=1, gpu_per_process=0),
        yt_client=yt_client,
        docker_image=DOCKER_IMAGE,
    )


def test_cypress_bind_file(yt_instance: YtInstance, yt_path: str, cypress_file: str) -> None:
    yt_client = yt_instance.get_client()

    destination_path = "bar"

    run(
        get_file_checker(destination_path, check_executable=True),
        backend=GenericBackend(),
        yt_path=yt_path,
        binds_cypress=[
            BindCypress(
                source=cypress_file,
                destination=destination_path,
            ),
        ],
        mesh=Mesh(node_count=1, process_per_node=1, gpu_per_process=0),
        yt_client=yt_client,
        docker_image=DOCKER_IMAGE,
    )


def test_cypress_bind_from_run_config(yt_instance: YtInstance, yt_path: str, cypress_file: str) -> None:
    yt_client = yt_instance.get_client()

    run_config = {
        "mesh": {
            "node_count": 1,
            "process_per_node": 1,
            "gpu_per_process": 0,
        },
        "bind_cypress": [
            f"{cypress_file}:bar",
        ],
    }

    with run_config_file(run_config) as run_config_path:
        tracto_cli = TractoCli(
            command=["python3", "/tractorun_tests/check_cypress_bind.py"],
            args=[
                "--run-config-path",
                run_config_path,
                "--yt-path",
                yt_path,
                "--bind-local",
                f"{get_data_path('../data/check_cypress_bind.py')}:/tractorun_tests/check_cypress_bind.py",
                "--user-config",
                json.dumps({"CYPRESS_FILE_TO_CHECK": "./bar"}),
            ],
        )
        op_run = tracto_cli.run()

    assert op_run.is_exitcode_valid()
    assert op_run.is_operation_state_valid(yt_client=yt_client, job_count=1)



import attrs

from tractorun.backend.tractorax.environment import Environment
from tractorun.base_backend import BackendBase


@attrs.define
class Tractorax(BackendBase):
    @property
    def environment(self) -> Environment:
        return Environment()



import uuid

from tests.utils import (
    DOCKER_IMAGE,
    make_cli_args,
    make_run_config,
    run_config_file,
)
from tractorun.backend.generic import GenericBackend
from tractorun.cli.tractorun_runner import make_configuration
from tractorun.mesh import Mesh
from tractorun.run import run
from tractorun.toolbox import Toolbox


def test_configuration() -> None:
    _, _, config = make_configuration(make_cli_args("--title", "cli title"))
    assert config.title == "cli title"

    run_config = make_run_config({"title": "config title"})
    with run_config_file(run_config) as run_config_path:
        _, _, config = make_configuration(["--run-config-path", run_config_path])
    assert config.title == "config title"

    with run_config_file(run_config) as run_config_path:
        _, _, config = make_configuration(["--run-config-path", run_config_path, "--title", "cli title"])
    assert config.title == "cli title"


def test_run_pickle(yt_path: str) -> None:
    def checker(toolbox: Toolbox) -> None:
        pass

    title = f"super title {uuid.uuid4()}"

    run_info = run(
        checker,
        yt_path=yt_path,
        title=title,
        mesh=Mesh(node_count=1, process_per_node=1, gpu_per_process=0),
        backend=GenericBackend(),
        docker_image=DOCKER_IMAGE,
        dry_run=True,
    )
    assert run_info.operation_spec["title"] == title



import importlib
import inspect
import pkgutil
from types import ModuleType
from typing import Iterable


MODULE_BLACKLIST = {
    # We have to ignore jax-related modules here
    # because jaxlib requires avx support, but orbstack doesn't have it.
    # Maybe we should build jaxlib from source for local test container.
    "tractorun.backend.tractorax",
    "tractorun.backend.tractorax.environment",
    "tractorun.backend.tractorax.backend",
}


def test_no_private_objects_in_public_namespace() -> None:
    # this test can't check variables with primitive types like strings, int, float and so on
    invalid_public_modules: list[tuple[str, object, ModuleType]] = []
    for module in _get_all_modules():
        if _is_private_module(module):
            continue
        for name, obj in inspect.getmembers(module):
            if name.startswith("_"):
                continue
            if not isinstance(obj, object):
                continue
            if inspect.ismodule(obj):
                continue
            obj_module = inspect.getmodule(obj)
            if obj_module is None:
                continue
            if not _is_private_module(obj_module):
                continue
            invalid_public_modules.append((name, obj, module))
    assert invalid_public_modules == []


def _get_all_modules() -> Iterable[ModuleType]:
    import tractorun

    for module_info in pkgutil.walk_packages(tractorun.__path__, tractorun.__name__ + "."):
        if module_info.name in MODULE_BLACKLIST:
            continue
        print(module_info.name, module_info.name in MODULE_BLACKLIST)
        yield importlib.import_module(module_info.name)


def _is_private_module(module: ModuleType | None) -> bool:
    if module is None:
        return False
    private_prefixes = [
        "tractorun.private",
        "tractorun.tests",
        "tractorun.cli",
    ]
    for prefix in private_prefixes:
        if module.__name__.startswith(prefix):
            return True
    return False
